//
//  BRStreamer.m
//  AudioPlayer
//
//  Created by Denny Kwon on 12/26/12.
//  Copyright (c) 2012 com.thegridmedia. All rights reserved.
//


#import "BRStreamer.h"

void interruptionListener(void *inClientData, UInt32 inInterruptionState)
{
	NSLog(@"STREAMER - interruptionListener");
	BRStreamer *self = (BRStreamer *)inClientData;
    [self pause];
	[[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:@"InterruptionNotification" object:nil]];
}

void AQOutputCallback(void *inUserData, AudioQueueRef inAQ, AudioQueueBufferRef inBuffer)
{
    /*
     This callback is mainly to used to track used buffers and recycle them after use.
	 This is where most of the threading locks are removed etc.
	 For now, don't worry about this because we are not recycling buffers.
	 Instead, we just create new buffers in memory whenever data gets called in from server */
	
	BRStreamer *self = (BRStreamer *)inUserData;
    unsigned long p = inBuffer->mAudioDataByteSize;
    self.totalBytesPlayed += p;
	AudioQueueFreeBuffer(self.queue, inBuffer); //since we're not recycling buffers, you need to free them after use every time.
	self.callbackCount++;
    
	NSLog(@"STREAMER - AQOutputCallback: bufferCount=%d, callbackCount=%d, %lu total bytes played", self.bufferCount, self.callbackCount, self.totalBytesPlayed);
    [self.source estimateDuration:self.totalBytesPlayed];
    
	if ((self.bufferCount-self.callbackCount) <= 1){
        if (self.flushing==FALSE){
            NSLog(@"STREAMER - AQOutputCallback: SEND MORE DATA!");
            [self parseBytes:[self.source getData]]; //this invokes the packet callback
        }
	}
    
	if ((self.bufferCount>1) && (self.bufferCount == self.callbackCount)){
        if (self.flushing==FALSE)
            [self finished];
	}
    
}


void propertyChangeCallback(void *data, AudioQueueRef inAQ, AudioQueuePropertyID inID)
{
    NSLog(@"STREAMER - propertyChangeCallback:");
	if (inID == kAudioQueueProperty_IsRunning){
        BRStreamer *self = (BRStreamer *)data;
        self.isRunning = YES;
		NSLog(@"STREAMER - propertyChangeCallback: kAudioQueueProperty_IsRunning"); //1634824814
	}
    else{ NSLog(@"STREAMER - propertyChangeCallback: not running"); }
}


//String reference: http://developer.apple.com/mac/library/documentation/Cocoa/Conceptual/Strings/Articles/formatSpecifiers.html
void propertyListenerCallback(void *inClientData, AudioFileStreamID inAudioFileStream, AudioFileStreamPropertyID inPropertyID, UInt32 *ioFlags)
{
	NSUInteger propID = (NSUInteger)inPropertyID;
	NSLog(@"STREAMER - propertyListenerCallback: %i", propID);
	BRStreamer *self = (BRStreamer *)inClientData;
    
    if (inPropertyID == kAudioFileStreamProperty_DataOffset){
        NSLog(@"STREAMER - propertyListenerCallback: kAudioFileStreamProperty_DataOffset");
        SInt64 offset;
        UInt32 offsetSize = sizeof(offset);
        OSStatus err = AudioFileStreamGetProperty(inAudioFileStream, kAudioFileStreamProperty_DataOffset, &offsetSize, &offset);
        if (err) {
            return;
        }
        self.source.dataOffset = offset;
    }
    
    if (inPropertyID != kAudioFileStreamProperty_ReadyToProducePackets){
        NSLog(@"STREAMER - not ready yet: %lu", inPropertyID);
        return;
    }
    
    
    
    NSLog(@"ALL SET TO GO! %lu", inPropertyID); //all metadata of the file is acquired
    
    // here we are getting the audio format and putting it into the 'asbd' struct:
    UInt32 dataSize = sizeof(AudioStreamBasicDescription);
    AudioStreamBasicDescription asbd; //this is a struct that holds info about the file
    AudioFileStreamGetProperty(inAudioFileStream, kAudioFileStreamProperty_DataFormat, &dataSize, &asbd);
    self.source.asbd = asbd;
    self.source.format = (self.source.asbd.mFormatID==kAudioFormatMPEG4AAC) ? FileFormatM4A : FileFormatMP3;

    // open up a new queue and set it to self's queue property:
    AudioQueueRef playQueue;
    AudioQueueNewOutput(&asbd, AQOutputCallback, self, CFRunLoopGetCurrent(), kCFRunLoopCommonModes, 0, &playQueue);
    self.queue = playQueue;
    AudioQueueAddPropertyListener(self.queue, kAudioQueueProperty_IsRunning, propertyChangeCallback, self); //add property listener
    

}

void packetCallback(void *inClientData, UInt32 inNumberBytes, UInt32 inNumberPackets, const void *inInputData, AudioStreamPacketDescription *inPacketDescriptions)
{
	NSLog(@"STREAMER - packetCallback: %lu packets, %lu bytes", inNumberPackets, inNumberBytes);
	BRStreamer *self = (BRStreamer *)inClientData;
	
	AudioQueueBufferRef bufferRef;
	AudioQueueAllocateBufferWithPacketDescriptions(self.queue, inNumberBytes, inNumberPackets, &bufferRef); //allocate a buffer the size of the received bytes
	
	memcpy(bufferRef->mAudioData, inInputData, inNumberBytes);
	bufferRef->mAudioDataByteSize = inNumberBytes; // before you added this line, the buffer wouldn't enqueue
	
	memcpy(bufferRef->mPacketDescriptions, inPacketDescriptions, sizeof(AudioStreamPacketDescription)*inNumberPackets);
	bufferRef->mPacketDescriptionCount = inNumberPackets;
    
	if (self.readyToAcceptPackets==TRUE){
		OSStatus status = AudioQueueEnqueueBuffer(self.queue, bufferRef, 0, NULL);
		if (status == noErr){
			self.bufferCount++;
            self.flushing = FALSE;
			NSLog(@"STREAMER - packetCallback: BUFFER ENQUEUED: %d", self.bufferCount);
        }
		else {
            NSLog(@"STREAMER - packetCallback: COULD NOT ENQUEUE BUFFER!");
        }
	}
	
	if (self.isRunning==FALSE){
		OSStatus status = AudioQueueStart(self.queue, NULL);
		if (status == noErr){
			self.isRunning = TRUE;
			NSLog(@"STREAMER - packetCallback: QUEUE STARTED");
		}
		else { NSLog(@"STREAMER - packetCallback: COULDN'T START THE QUEUE!"); }
	}
    
    for (int i = 0; i < inNumberPackets; ++i){
        self.processedPacketsCount++;
        self.processedPacketsSizeTotal += inPacketDescriptions[i].mDataByteSize;
    }
    
}




#define kUpdateFreq 0.50f
#define BitRateEstimationMinPackets 50

@interface BRStreamer ()
@property (nonatomic) BOOL falseFinish;
@property (nonatomic) BOOL suspended;
@end

@implementation BRStreamer
@synthesize fileStream;
@synthesize queue;
@synthesize isRunning;
@synthesize bufferCount;
@synthesize callbackCount;
@synthesize delegate;
@synthesize readyToAcceptPackets;
@synthesize flushing;
@synthesize source;
@synthesize tracks;
@synthesize timer;
@synthesize processedPacketsCount;		// number of packets accumulated for bitrate estimation
@synthesize processedPacketsSizeTotal;	// byte size of accumulated estimation packets
@synthesize totalBytesPlayed;
@synthesize index;

- (id)init
{
	NSLog(@"BRStreamer - init");
	self = [super init];
	if (self){
        self.index = 0;
        self.processedPacketsCount = 0;
        self.processedPacketsSizeTotal = 0;
        self.totalBytesPlayed = 0;
        self.falseFinish = FALSE;
        self.flushing = FALSE;
		self.isRunning = FALSE;
		self.readyToAcceptPackets = FALSE;
        self.suspended = NO;
        
        AudioSessionInitialize(NULL, NULL, interruptionListener, self);
        UInt32 sessionCategory = kAudioSessionCategory_MediaPlayback; 
        AudioSessionSetProperty(kAudioSessionProperty_AudioCategory, sizeof(sessionCategory), &sessionCategory);
	}
	return self;
}

+ (void)initialize
{
	NSLog(@"BRStreamer: initialize");
}


+ (BRStreamer *)streamer
{
    BRStreamer *p = [[BRStreamer alloc] init];
    return [p autorelease];
}


- (void)dealloc
{
	NSLog(@"BRStreamer - DEALLOC");
    if (self.timer)
        [self.timer invalidate];
    
    self.timer = nil;
    AudioQueueStop(self.queue, false); //calling 'false' here flushes the AudioQueue so you don't have to call 'Flush' later.  it's also a smoother transition
	AudioQueueDispose(self.queue, true);
	AudioFileStreamClose(self.fileStream);
    
	self.fileStream = nil;
	self.queue = nil;
	self.delegate = nil;
	AudioSessionSetActive(false);
	if (source){
        [source release];
        source = nil;
    }
    self.tracks = nil;
    [super dealloc];
}


//this is only called when starting a new stream NOT when resuming from a pause.
//resuming from a pause is done with: - (void)resume
- (void)start
{
    int i = self.index % self.tracks.count;
    self.index++;
    BRTrack *track = (BRTrack *)[self.tracks objectAtIndex:i];
    if (!track)
        return;
    
    [track checkCache];
    self.source = track;
	NSLog(@"BRStreamer - start: %@, index==%@", self.source.name, self.source.fileIndex);
    
	AudioSessionSetActive(true);
    AudioFileTypeID fileType = (self.source.format==FileFormatM4A) ? kAudioFileMPEG4Type : kAudioFileMP3Type;
    
    OSStatus status = AudioFileStreamOpen(self, propertyListenerCallback, packetCallback, fileType, &fileStream); //open the file
    if (status)
        NSLog(@"ERROR: AudioFileStreamOpen = %ld", status);

    UInt32 sessionCategory = kAudioSessionCategory_MediaPlayback;
	AudioSessionSetProperty(kAudioSessionProperty_AudioCategory, sizeof(sessionCategory), &sessionCategory);
    
	self.bufferCount = 0;
    self.callbackCount = 0;
	self.readyToAcceptPackets = TRUE;

    self.source.delegate = self;
    
    if (self.source.full){
        [self startPlayback];
        return;
    }
    
    if (self.source.fileData){ // downloading in progress
        if (self.source.fileData.length > [self.source minDataSize]) //enough bytes already downloaded, start playback right away
            [self startPlayback];

        return;
    }
    
    if (self.source.connection==nil){ // start from scratch
        [self.source download];
        return;
    }
    
}


- (void)pause
{
	NSLog(@"Streamer: pause");
    if (self.timer!=nil)
        [self.timer invalidate];
    
    self.timer = nil;
    self.isRunning = NO;
	AudioQueuePause(queue);
}

- (void)skip
{
    [self pause]; //pause current track
    [self queueNext];
}

- (void)resume
{
	NSLog(@"Streamer: resume");
    if (self.suspended){
        [self start];
        self.suspended = NO;
        return;
    }
    
    self.isRunning = TRUE;
	AudioQueueStart(queue, NULL);
    if (self.timer==nil)
        self.timer = [NSTimer scheduledTimerWithTimeInterval:kUpdateFreq target:self selector:@selector(playBackTimer) userInfo:nil repeats:YES];
    
    if (self.falseFinish)
        return;
    
    [self parseBytes:[self.source getData]];
    self.falseFinish = NO;
}

- (void)suspend
{
    //this is called when recorder is started because only one audio session should be active at at time. before doing this, tracks wouldn't record after playing previously
    NSLog(@"STREAMER - SUSPEND");
    [self clear];
    AudioQueueStop(self.queue, false); //calling 'false' here flushes the AudioQueue so you don't have to call 'Flush' later.  it's also a smoother transition
	AudioQueueDispose(self.queue, true);
	AudioFileStreamClose(self.fileStream);
	AudioSessionSetActive(false);
    self.suspended = YES;
    self.index--;
}

- (void)clear
{
    NSLog(@"STREAMER - CLEAR");
    self.isRunning = NO;
    self.flushing = YES;
    [self pause];
    self.source.delegate = nil;
    [self.source reset];
}

- (void)queueNext
{
    [self clear]; //clear everything out.
    [self reset:TRUE];
}

- (void)finished
{
    NSLog(@"STREAMER - FINISHED");
    if (source.connection != nil){
        
        NSLog(@"STERAMER: FALSE FINISH - Ran out of data");
        self.falseFinish = TRUE;
        [self pause];
        [self performSelector:@selector(resume) withObject:nil afterDelay:5.0f];
        return;
    }
    
    NSLog(@"STERAMER: Conventional Finish");
    [[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:kBRStreamerFinishedNotification object:nil]];
    if (self.timer){
        [self.timer invalidate];
        self.timer = nil;
    }

    [self.source reset];
    if ([self.delegate respondsToSelector:@selector(fileFinished)])
         [self.delegate fileFinished];
    
    //Have to add some buffer time here because the last few bytes need to play out before clearing out the player and starting the next file.
    [self performSelector:@selector(queueNext) withObject:nil afterDelay:1.0f];
}

- (void)parseBytes:(NSData *)data
{
//	OSStatus status = AudioFileStreamParseBytes(self.fileStream, data.length, data.bytes, kAudioFileStreamParseFlag_Discontinuity); // ! this caused buffering issues on MP3s - on packetCallback, only one buffer was loaded at a time which led to noticeable skips.
    
	OSStatus status = AudioFileStreamParseBytes(self.fileStream, data.length, data.bytes, 0);
    NSLog(@"STREAMER - parse bytes: %ld", status); // status=1 means error, 0 means fine.
}

- (void)reset:(BOOL)restart
{
	//this basically just resets everything.  kills the queue, closes the file stream, and restarts everything by calling [self play]
	NSLog(@"STREAMER - reset");
    if (self.isRunning){
        [[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:kBRStreamerFinishedNotification object:nil]];
        self.flushing = YES;
    }
    if (self.timer){
        [self.timer invalidate];
        self.timer = nil;
    }
    
	AudioQueueStop(self.queue, false); //calling 'false' here flushes the AudioQueue so you don't have to call 'Flush' later.  it's also a smoother transition
	AudioQueueDispose(self.queue, true);
	AudioFileStreamClose(self.fileStream);
	self.isRunning = FALSE;
	self.readyToAcceptPackets = FALSE;
    self.processedPacketsCount = 0;
    self.processedPacketsSizeTotal = 0;
    self.totalBytesPlayed = 0;
    
    self.flushing = NO;
    
    if (self.source)
        [self.source reset];
    
    if (restart==TRUE)
        [self start];
    
}

/* newSeekTime is NOT seconds, it is units of kUpdateFreq. */
- (void)internalSeekToTime:(double)newSeekTime
{
    NSLog(@"internalSeekToTime:");
    double calculatedBitRate = [self.source calculatedBitRate:self.processedPacketsSizeTotal packetsCount:self.processedPacketsCount];
    if (calculatedBitRate == 0.0 ){
//        NSLog(@"internalSeekToTime: TEST 1");
		return;
	}
    
    /* http://www.cocoawithlove.com/2010/03/streaming-mp3aac-audio-again.html
     "However, it is not safe (particularly in an AAC) to seek to an arbitrary location. You must seek to a packet boundary."  */
    
	// Calculate the byte offset for seeking
    NSInteger offset = self.source.dataOffset;
	NSInteger seekByteOffset = offset + (newSeekTime / self.source.duration) * (self.source.fileData.length - offset);

    
    /* Attempt to leave 1 useful packet at the end of the file (although in
	   reality, this may still seek too far if the file has a long trailer). */
    
//	if (seekByteOffset > self.source.fileData.length - 2 * packetBufferSize){
//		seekByteOffset = self.source.fileData.length - 2 * packetBufferSize;
//	}

	// Attempt to align the seek with a packet boundary:
//    NSLog(@"internalSeekToTime: TEST 2");
	double packetDuration = self.source.asbd.mFramesPerPacket / self.source.asbd.mSampleRate;
	calculatedBitRate = [self.source calculatedBitRate:self.processedPacketsSizeTotal packetsCount:self.processedPacketsCount];
    
    if (packetDuration==0){
//        NSLog(@"internalSeekToTime: ERROR - packetDuration==0");
        return;
    }
    
    if (calculatedBitRate==0){
//        NSLog(@"internalSeekToTime: ERROR - calculatedBitRate==0");
        return;
    }
    
    UInt32 ioFlags = 0;
    SInt64 packetAlignedByteOffset;
    SInt64 seekPacket = floor(newSeekTime / packetDuration);
    OSStatus err = AudioFileStreamSeek(self.fileStream, seekPacket, &packetAlignedByteOffset, &ioFlags);
    if (err){
//        NSLog(@"internalSeekToTime: TEST 5");
        return;
    }
    if ((ioFlags & kAudioFileStreamSeekFlag_OffsetIsEstimated)==true){
//        NSLog(@"internalSeekToTime: TEST 7");
        return;
    }
    
//    NSLog(@"internalSeekToTime: TEST 7: BYTE OFFSET == %.2d", seekByteOffset);
    seekByteOffset = packetAlignedByteOffset + offset;
    [self jump:seekByteOffset time:newSeekTime];
}

- (void)seekto:(double)value
{
    if (self.source.duration==0)
        self.source.duration = [self.source estimateDuration:self.totalBytesPlayed]; //get time estimate
    
    double estimate = self.source.duration;
    double seekTo = value * estimate;

    if (self.source.format==FileFormatM4A){ // seek for M4A
        NSLog(@"STREAMER - ESTIMATE: %.2f, SEEK TO: %.2f", estimate, seekTo);
        [self internalSeekToTime:seekTo];
        return;
    }
    
    // seek for MP3:
    double off = (value * self.source.fileSize);
    [self jump:off time:seekTo];
}

/*
- (void)jump:(double)byteOffset time:(double)seekTime
{
    NSLog(@"jump: %.2f", byteOffset);
    if (byteOffset > self.source.fileData.length){ // show alert
        
        //prepare request for data from offset
        [self pause];
        AudioQueueFlush(self.queue);
        if (self.source.connection){
            [self.source.connection cancel];
            self.source.connection.delegate = nil;
            self.source.connection = nil;
        }
        
        self.source.fileData = [NSMutableData data];
        self.source.progressIndex = seekTime;
        self.source.leftoff = 0;
        self.totalBytesPlayed = 0;
        int mb = 1024*1024;
        [self.source download:NSMakeRange(20*mb, 10*mb)];
        self.source.delegate = self;
        // show loading indicator

        
        return;
    }

    self.source.progressIndex = seekTime;
    self.source.leftoff = byteOffset;
    self.totalBytesPlayed = byteOffset;


    self.flushing = TRUE;
    [self pause]; //this kills the timer too.
    AudioQueueReset(self.queue);
    self.flushing = NO;
    [self startPlayback];
}
 */

- (void)jump:(double)byteOffset time:(double)seekTime
{
    NSLog(@"jump: %.2f", byteOffset);
    if (byteOffset > self.source.fileData.length){ // show alert
        [self resume];
        return;
    }
    
    self.source.progressIndex = seekTime;
    self.source.leftoff = byteOffset;
    self.totalBytesPlayed = byteOffset;
    
    
    self.flushing = TRUE;
    [self pause]; //this kills the timer too.
    AudioQueueReset(self.queue);
    self.flushing = NO;
    [self startPlayback];
}

- (void)playBackTimer
{
    [[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:kBRStreamerPlaybackNotification object:nil]];
    
    NSString *running = @"YES";
    if (self.isRunning==FALSE){ //some files need more data to be pushed in before they start playing. This covers for that.
        running = @"NO";
        [self parseBytes:[source getData]];
    }
    
    NSLog(@"STREAMER - playBackTimer: RUNNING = %@", running);
    
    double pct = [self.source incrementIndex:kUpdateFreq];
    double duration = (self.source.duration > 0) ?  self.source.duration : [self.source estimateDuration:self.totalBytesPlayed];
    double progress = self.source.progressIndex / duration;
    [self.delegate updateProgress:progress];
    
    if (pct > 0.7){
        int i = [self.source.fileIndex intValue]+1;
        i = i % self.tracks.count;
        BRTrack *nextTrack = (BRTrack *)[self.tracks objectAtIndex:i];
        NSLog(@"BR STREMAER - ALMOST DONE, Next Track == %@", nextTrack.name);
        if (!nextTrack.fileData)
            [nextTrack prepare];
    }
    
    if ([running isEqualToString:@"YES"])
        return;
    

}

#pragma mark - BRTrackDelegate
- (void)startPlayback //delegate callback from BRTrack - called when enough bytes are loaded either from server or file:
{
    NSLog(@"BR STREAMER - startPlayback");
    [[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:kBRStreamerStartedNotification object:nil]];

    [self parseBytes:[self.source getData]];
    if (self.timer)
        return;
    
    self.timer = [NSTimer scheduledTimerWithTimeInterval:kUpdateFreq target:self selector:@selector(playBackTimer) userInfo:nil repeats:YES];
    [self.delegate playbackStarted];
}




@end
